//--------------------------------------------------------------------------
// biyon1 (0.1)
//--------------------------------------------------------------------------
#include <stdio.h>
#include <windows.h>
#include <math.h>



#define SND_SAMPLES        (44100)
#define SND_CH             (2)
#define SND_BITS           (16)
#define SND_SEC            (7)
#define SND_PI             (3.14159265358979323846f)
#define SND_BASE           (6.0f / SND_SAMPLES)
#define SND_DT             (1.05946309435929526456f) //root12(2) = 1.0594630943592952645618252949463
#define SND_LENGTH         (SND_SEC * SND_SAMPLES * SND_CH)
#define SND_LENGTH_BYTE    (SND_LENGTH * (SND_BITS / 8))

static HWAVEOUT     hWOut = NULL;
static WAVEFORMATEX wfx;
static WAVEHDR      hdr;
static float buffer[SND_LENGTH * 2];

float f_rand() {
  static int a = 1;
  static int b = 234567;
  static int c = 890123;
  a += b; b += c; c += a;
  return (float)(a >> 16) * (1.0f / 32768.0f);
}

//Make WAVE format
WAVEFORMATEX mk_wav_fmt(int samples, int channel, int bit) {
  WAVEFORMATEX ret = { 
    WAVE_FORMAT_PCM, channel, samples, 0, 0, bit, sizeof(WAVEFORMATEX)
  };
  ret.nBlockAlign     = (ret.nChannels * ret.wBitsPerSample) / 8;
  ret.nAvgBytesPerSec = ret.nSamplesPerSec * ret.nBlockAlign;
  return ret;
}

//Make WAVE header
WAVEHDR mk_wav_hdr(LPSTR buffer, int len) {
  WAVEHDR ret = {
    buffer, len, 0, 0, 0, 0, NULL, NULL,
  };
  return ret;
}

//float32 -> short
void snd_convert_buf(float *buf, int len, float gain) {
  short *temp = (short *)buf;
  for(int i = 0 ; i < len; i++) {
    float k = buf[i] * gain;
    //temp[i] = (short)(k * (float)32767.0f);
    temp[i] = (short)(sin(k) * (float)32767.0f);
  }
}

void snd_term() {
  waveOutReset(hWOut);
  waveOutUnprepareHeader(hWOut, &hdr, sizeof(WAVEHDR));
  waveOutClose(hWOut);
}


//------------------------------------------------------------
//render splash sound.
//------------------------------------------------------------

double snd_round(float x) {
  __asm {
    fld  dword ptr [x]
    frndint
  }
}


float snd_gen_dt(int n) {
  float t = SND_BASE;
  while(n-- > 0) t *= SND_DT;
  return t;
}


float snd_saw(float t, float m) {
  float g = t * m;
  return (g - snd_round(g)) * 2;
}

float snd_sin(float t, float m) {
  return sin( (t * m) * SND_PI * 2);
}

float snd_sqr(float t, float m) {
  float g = snd_saw(t, m);
  return (g > 0.25) ? 1 : -1;
}

void snd_delay(int dt, float gain) {
  int phase = dt;
  for(int i = 0 ; i < SND_LENGTH; i += SND_CH) {
    phase--;
    if(phase < 0) {
      phase += dt;
    }
    int g = phase * 2;
    float a = buffer[g + 0] * gain + buffer[i + 0];
    float b = buffer[g + 1] * gain + buffer[i + 1];
    buffer[i + 0] = buffer[g + 0] = a;
    buffer[i + 1] = buffer[g + 1] = b;
  }
}


//render splash sound.
int gg = 10;
float (*func_osc[3])(float, float) = {
  snd_saw,
  snd_sin,
  snd_sqr,
};

struct tone_data {
  int    osc;
  float  detune;
  float  fm;
  float  noise;
};

tone_data  cdata[2] = {
  //{0, 1.000, 0.0, 0.0},
  //{0, 1.000, 0.0, 0.0},
  {1, 1.010000, 2.500000, 0.000000 },
  {1, 0.997000, 2.799999, 0.000000 },
};

void snd_render() {
  double t = 0;
  double dt = snd_gen_dt(gg + 36);
  ZeroMemory(buffer, sizeof(buffer));
  printf("Tone = %d\n", gg);
  float  vol = 1;
  float dvol = 0.99997;
  
    for(int i = 0 ; i < SND_LENGTH; i += SND_CH) {
      float v = 0;
      for(int ch = 0; ch < 2; ch++) {
        tone_data *w = &(cdata[ch]);
        float (*f_osc)(float, float) = func_osc[w->osc];
        float g = f_osc(t, 1) + f_osc(t, w->detune);
        if(w->fm > 0) {
          g = sin(g * w->fm * SND_PI);
        }
        v += g;
        v += f_rand() * w->noise;
      }
      
      buffer[i + 0] += v * vol;
      buffer[i + 1] += v * vol;
      vol *= dvol;
      t += dt;
  }
  
  //Convert and render
  snd_convert_buf(buffer, SND_LENGTH, 0.3);
  wfx   = mk_wav_fmt(SND_SAMPLES, SND_CH, SND_BITS);
  hdr   = mk_wav_hdr((LPSTR)(&buffer[0]), SND_LENGTH_BYTE);

  waveOutOpen(&hWOut, WAVE_MAPPER, &wfx, 0, 0, CALLBACK_NULL);
  waveOutPrepareHeader(hWOut , &hdr , sizeof(WAVEHDR));
  waveOutWrite(hWOut, &hdr , sizeof(WAVEHDR));
}



void snd_save(char *outname) {
  int    fmtChunkBytes  = SND_BITS;
  short  audioFormat    = 1;
  short  numChannels    = SND_CH;
  int    sampleRate     = SND_SAMPLES;
  short  bitsPerSample  = SND_BITS;
  int    byteRate       = bitsPerSample * numChannels / 8 * sampleRate;
  short  blockAlign     = bitsPerSample * numChannels / 8;
  int    dataChunkBytes = SND_LENGTH_BYTE;// * bitsPerSample * numChannels / 8;
  int    total          = 44 + dataChunkBytes;

  OFSTRUCT ofs;
  DWORD fc;
  HANDLE h = LongToHandle(OpenFile(outname, &ofs, OF_CREATE));
  WriteFile(h, "RIFF"          , 4, &fc, 0);
  WriteFile(h, &total          , 4, &fc, 0);
  WriteFile(h, "WAVEfmt "      , 8, &fc, 0);
  WriteFile(h, &fmtChunkBytes  , 4, &fc, 0);
  WriteFile(h, &audioFormat    , 2, &fc, 0);
  WriteFile(h, &numChannels    , 2, &fc, 0);
  WriteFile(h, &sampleRate     , 4, &fc, 0);
  WriteFile(h, &byteRate       , 4, &fc, 0);
  WriteFile(h, &blockAlign     , 2, &fc, 0);
  WriteFile(h, &bitsPerSample  , 2, &fc, 0);
  WriteFile(h, "data"          , 4, &fc, 0);
  WriteFile(h, &dataChunkBytes , 4, &fc, 0);
  WriteFile(h, &buffer[0]  , dataChunkBytes, &fc,0);
  CloseHandle(h);
}


static int stick = 0;
void disp_param() {
  system("cls");
  printf("************************************************************\n");
  printf("   т[ v0.1\n");
  printf("************************************************************\n");
  printf("\n");
  for(int i = 0 ; i < 2; i++) {
    char *str[] = {
      "SAW",
      "SIN",
      "SQR",
    };
    tone_data *w = &(cdata[i]);
    if(stick == i) printf("-->");
    else           printf("   ");
    printf("osc[%02d]   = %s, detune = %f, FM = %f,   Noize=%f\n", i, str[w->osc], w->detune, w->fm, w->noise);
  }
  printf("\n");
  printf("\n");
  printf("PITCH     = %02d\n", gg);
  printf("STICK     = %02d\n", stick);
  printf("\n");
  printf("OSC       = A=saw S=sin D=sqr\n");
  printf("Detune    = Q -,    W +\n");
  printf("FM        = E -,    R +\n");
  printf("NOISE     = T -,    Y +\n");
  printf("Pitch     = O -,    P +\n");
  printf("Stick     = 1,      2\n");
  
  printf("\n");
  printf("SaveWave  : F7\n");
  printf("SaveParam : F8\n");
  printf("SpaceKey  : Prev Sound\n");
}

void save_param(char *name) {
  FILE *fp = fopen(name, "w");
  if(fp) {
    for(int i = 0 ; i < 2; i++) {
      tone_data *w = &(cdata[i]);
      fprintf(fp, "{%d, %f, %f, %f },\n", w->osc, w->detune, w->fm, w->noise);
    }
    fclose(fp);
  }
}


int PollKey(int x){
  if(GetAsyncKeyState(x) & 0x8000) {
    disp_param();
    return 1;
  }
  return 0;
}


int main() {
  snd_render();
  disp_param();
  while(!PollKey(VK_ESCAPE)) {
    if(PollKey(VK_F5) || PollKey(VK_SPACE)) {
      snd_term();
      snd_render();
    }
    if(PollKey(VK_F7)) {
      snd_save("out.wav");
    }
    
    if(PollKey(VK_F8)) {
      save_param("import.txt");
    }
    
    tone_data *w = &(cdata[stick]);
    
    //Pitch
    if(PollKey('O')) {gg--;    }
    if(PollKey('P')) {gg++;    }
    
    //Stick Tone
    if(PollKey('1')) {stick     = 0;    }
    if(PollKey('2')) {stick     = 1;    }
    
    //Wave
    if(PollKey('A')) {w->osc    = 0;    }
    if(PollKey('S')) {w->osc    = 1;    }
    if(PollKey('D')) {w->osc    = 2;    }
    
    //Detune
    if(PollKey('Q')) {w->detune -= 0.001;    }
    if(PollKey('W')) {w->detune += 0.001;    }

    //FM
    if(PollKey('E')) {w->fm -= 0.1;    }
    if(PollKey('R')) {w->fm += 0.1;    }
    
    //Noise
    if(PollKey('T')) {w->noise -= 0.01;    }
    if(PollKey('Y')) {w->noise += 0.01;    }

    Sleep(16);
    
  }
  snd_term();
}


